# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Citation system for NetKet.

This module provides tools for citation tracking and bibliography generation.
"""

__all__ = [
    "cite",
    "reference",
    "get_all_citations",
    "validate_references",
    "get_citations_for_docs",
]


from netket.utils.citations.core import (
    cite,
    reference,
    get_all_citations,
    validate_references,
)
from netket.utils.citations.docs import (
    get_citations_for_docs,
)
